@@ -1,7 +1,6 @@
 /*
  * Copyright 2005-2012 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/paxinc.h,v 1.17 2012/11/04 07:26:24 vapier Exp $
  *
  * Copyright 2005-2012 Ned Ludd        - <solar@gentoo.org>
  * Copyright 2005-2012 Mike Frysinger  - <vapier@gentoo.org>
@@ -16,7 +15,10 @@
 #include "xfuncs.h"
 
 #ifndef VERSION
-# define VERSION "cvs"
+# define VERSION "git"
+#endif
+#ifndef VCSID
+# define VCSID "<unknown>"
 #endif
 
 /* ELF love */
