@@ -1,13 +1,11 @@
 /*
  * Copyright 2003-2012 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.269 2014/11/20 01:25:56 vapier Exp $
  *
  * Copyright 2003-2012 Ned Ludd        - <solar@gentoo.org>
  * Copyright 2004-2012 Mike Frysinger  - <vapier@gentoo.org>
  */
 
-static const char rcsid[] = "$Id: scanelf.c,v 1.269 2014/11/20 01:25:56 vapier Exp $";
 const char argv0[] = "scanelf";
 
 #include "paxinc.h"
@@ -189,13 +187,13 @@ static void scanelf_file_get_symtabs(elfobj *elf, void **sym, void **str)
 	Elf ## B ## _Shdr *edynsym = dynsym; \
 	Elf ## B ## _Shdr *edynstr = dynstr; \
 	\
-	if (symtab && EGET(esymtab->sh_type) == SHT_NOBITS) \
+	if (!VALID_SHDR(elf, esymtab)) \
 		symtab = NULL; \
-	if (dynsym && EGET(edynsym->sh_type) == SHT_NOBITS) \
+	if (!VALID_SHDR(elf, edynsym)) \
 		dynsym = NULL; \
-	if (strtab && EGET(estrtab->sh_type) == SHT_NOBITS) \
+	if (!VALID_SHDR(elf, estrtab)) \
 		strtab = NULL; \
-	if (dynstr && EGET(edynstr->sh_type) == SHT_NOBITS) \
+	if (!VALID_SHDR(elf, edynstr)) \
 		dynstr = NULL; \
 	\
 	/* Use the set with more symbols if both exist. */ \
@@ -229,6 +227,8 @@ static void scanelf_file_get_symtabs(elfobj *elf, void **sym, void **str)
 	 * reconstruct the section header info out of the dynamic
 	 * tags so we can see what symbols this guy uses at runtime.
 	 */
+	if (!elf->phdr)
+		return;
 #define GET_SYMTABS_DT(B) \
 	if (elf->elf_class == ELFCLASS ## B) { \
 	size_t i; \
@@ -273,27 +273,52 @@ static void scanelf_file_get_symtabs(elfobj *elf, void **sym, void **str)
 	for (i = 0; i < EGET(ehdr->e_phnum); i++) { \
 		Elf ## B ## _Addr vaddr = EGET(phdr[i].p_vaddr); \
 		Elf ## B ## _Addr filesz = EGET(phdr[i].p_filesz); \
-		offset = EGET(phdr[i].p_offset); \
+		Elf ## B ## _Off hash_offset = offset + (vhash - vaddr); \
 		\
 		if (EGET(phdr[i].p_type) != PT_LOAD) \
 			continue; \
 		\
+		offset = EGET(phdr[i].p_offset); \
+		if (offset >= (uint64_t)elf->len) \
+			goto corrupt_hash; \
+		if (filesz >= (uint64_t)elf->len) \
+			goto corrupt_hash; \
+		if (hash_offset + (sizeof(Elf32_Word) * 4) > (uint64_t)elf->len) \
+			goto corrupt_hash; \
+		\
 		if (vhash >= vaddr && vhash < vaddr + filesz) { \
 			/* Scan the hash table to see how many entries we have */ \
 			Elf32_Word max_sym_idx = 0; \
-			Elf32_Word *hashtbl = elf->vdata + offset + (vhash - vaddr); \
+			Elf32_Word *hashtbl = elf->vdata + hash_offset; \
 			Elf32_Word b, nbuckets = EGET(hashtbl[0]); \
 			Elf32_Word nchains = EGET(hashtbl[1]); \
 			Elf32_Word *buckets = &hashtbl[2]; \
 			Elf32_Word *chains = &buckets[nbuckets]; \
 			Elf32_Word sym_idx; \
+			Elf32_Word chained; \
+			\
+			if (hash_offset >= (uint64_t)elf->len) \
+				goto corrupt_hash; \
+			if (nbuckets >= UINT32_MAX / 4) \
+				goto corrupt_hash; \
+			if (nchains >= UINT32_MAX / 4) \
+				goto corrupt_hash; \
+			if (nbuckets * 4 > elf->len - offset) \
+				goto corrupt_hash; \
+			if (nchains * 4 > elf->len - offset) \
+				goto corrupt_hash; \
 			\
 			for (b = 0; b < nbuckets; ++b) { \
 				if (!buckets[b]) \
 					continue; \
-				for (sym_idx = buckets[b]; sym_idx < nchains && sym_idx; sym_idx = chains[sym_idx]) \
+				for (sym_idx = buckets[b], chained = 0; \
+				     sym_idx < nchains && sym_idx && chained <= nchains; \
+				     sym_idx = chains[sym_idx], ++chained) { \
 					if (max_sym_idx < sym_idx) \
 						max_sym_idx = sym_idx; \
+				} \
+				if (chained > nchains) \
+					goto corrupt_hash; \
 			} \
 			ESET(sym_shdr.sh_size, sym_shdr.sh_entsize * max_sym_idx); \
 		} \
@@ -311,6 +336,10 @@ static void scanelf_file_get_symtabs(elfobj *elf, void **sym, void **str)
 	}
 	GET_SYMTABS_DT(32)
 	GET_SYMTABS_DT(64)
+	return;
+
+ corrupt_hash:
+	warn("%s: ELF hash table is corrupt", elf->filename);
 }
 
 static char *scanelf_file_pax(elfobj *elf, char *found_pax)
@@ -430,9 +459,9 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 	} else if (elf->shdr != NULL) { \
 		/* no program headers which means this is prob an object file */ \
 		Elf ## B ## _Shdr *shdr = SHDR ## B (elf->shdr); \
-		Elf ## B ## _Shdr *strtbl = shdr + EGET(ehdr->e_shstrndx); \
-		char *str; \
-		if ((void*)strtbl > elf->data_end) \
+		uint16_t shstrndx = EGET(ehdr->e_shstrndx); \
+		Elf ## B ## _Shdr *strtbl = shdr + shstrndx; \
+		if (shstrndx >= elf->len - sizeof(*strtbl) || !VALID_SHDR(elf, strtbl)) \
 			goto skip_this_shdr##B; \
 		/* let's flag -w/+x object files since the final ELF will most likely \
 		 * need write access to the stack (who doesn't !?).  so the combined \
@@ -442,9 +471,9 @@ static char *scanelf_file_phdr(elfobj *elf, char *found_phdr, char *found_relro,
 		for (i = 0; i < EGET(ehdr->e_shnum); ++i) { \
 			if (EGET(shdr[i].sh_type) != SHT_PROGBITS) continue; \
 			offset = EGET(strtbl->sh_offset) + EGET(shdr[i].sh_name); \
-			str = elf->data + offset; \
-			if (str > elf->data + offset + sizeof(NOTE_GNU_STACK)) continue; \
-			if (!strcmp(str, NOTE_GNU_STACK)) { \
+			if (offset >= elf->len - sizeof(NOTE_GNU_STACK)) \
+				continue; \
+			if (!strcmp(elf->data + offset, NOTE_GNU_STACK)) { \
 				if (multi_stack++) warnf("%s: multiple .note.GNU-stack's !?", elf->filename); \
 				flags = EGET(shdr[i].sh_flags); \
 				if (be_quiet && ((flags & check_flags) != check_flags)) \
@@ -1068,37 +1097,51 @@ static const char *scanelf_file_needed_lib(elfobj *elf, char *found_needed, char
 }
 static char *scanelf_file_interp(elfobj *elf, char *found_interp)
 {
-	void *strtbl_void;
+	uint64_t offset = 0;
 
 	if (!show_interp) return NULL;
 
-	strtbl_void = elf_findsecbyname(elf, ".interp");
-
-	if (strtbl_void) {
-#define SHOW_INTERP(B) \
-		if (elf->elf_class == ELFCLASS ## B) { \
-			Elf ## B ## _Shdr *strtbl = SHDR ## B (strtbl_void); \
-			*found_interp = 1; \
-			return (be_wewy_wewy_quiet ? NULL : elf->data + EGET(strtbl->sh_offset)); \
-		}
-		SHOW_INTERP(32)
-		SHOW_INTERP(64)
-	} else {
+	if (elf->phdr) {
 		/* Walk all the program headers to find the PT_INTERP */
 #define SHOW_PT_INTERP(B) \
 		if (elf->elf_class == ELFCLASS ## B) { \
-		unsigned long i; \
-		Elf ## B ## _Ehdr *ehdr = EHDR ## B (elf->ehdr); \
-		Elf ## B ## _Phdr *phdr = PHDR ## B (elf->phdr); \
-		for (i = 0; i < EGET(ehdr->e_phnum); i++) { \
-			if (EGET(phdr[i].p_type) != PT_INTERP) \
-				continue; \
-			*found_interp = 1; \
-			return (be_wewy_wewy_quiet ? NULL : elf->data + EGET(phdr[i].p_offset)); \
-		} \
+			size_t i; \
+			Elf ## B ## _Ehdr *ehdr = EHDR ## B (elf->ehdr); \
+			Elf ## B ## _Phdr *phdr = PHDR ## B (elf->phdr); \
+			for (i = 0; i < EGET(ehdr->e_phnum); ++i) { \
+				if (EGET(phdr[i].p_type) == PT_INTERP) { \
+					offset = EGET(phdr[i].p_offset); \
+					break; \
+				} \
+			} \
 		}
 		SHOW_PT_INTERP(32)
 		SHOW_PT_INTERP(64)
+	} else if (elf->shdr) {
+		/* Use the section headers to find it */
+		void *strtbl_void = elf_findsecbyname(elf, ".interp");
+
+#define SHOW_INTERP(B) \
+		if (elf->elf_class == ELFCLASS ## B) { \
+			Elf ## B ## _Shdr *strtbl = SHDR ## B (strtbl_void); \
+			offset = EGET(strtbl->sh_offset); \
+		}
+		if (strtbl_void) {
+			SHOW_INTERP(32)
+			SHOW_INTERP(64)
+		}
+	}
+
+	/* Validate the pointer even if we don't use it in output */
+	if (offset && offset <= (uint64_t)elf->len) {
+		char *interp = elf->data + offset;
+
+		/* If it isn't a C pointer, it's garbage */
+		if (memchr(interp, 0, elf->len - offset)) {
+			*found_interp = 1;
+			if (!be_wewy_wewy_quiet)
+				return interp;
+		}
 	}
 
 	return NULL;
@@ -1381,29 +1424,28 @@ static const char *scanelf_file_sym(elfobj *elf, char *found_sym)
 		Elf ## B ## _Word i, cnt = EGET(symtab->sh_entsize); \
 		char *symname; \
 		size_t ret_len = 0; \
+		if (!VALID_SHDR(elf, symtab) || !VALID_SHDR(elf, strtab)) \
+			goto break_out; \
 		if (cnt) \
 			cnt = EGET(symtab->sh_size) / cnt; \
 		for (i = 0; i < cnt; ++i) { \
-			if ((void*)sym > elf->data_end) { \
-				warnf("%s: corrupt ELF symbols - aborting", elf->filename); \
+			if ((void *)sym >= elf->data_end - sizeof(*sym)) \
 				goto break_out;	\
-			} \
 			if (sym->st_name) { \
 				/* make sure the symbol name is in acceptable memory range */ \
 				symname = elf->data + EGET(strtab->sh_offset) + EGET(sym->st_name); \
-				if ((void*)symname > elf->data_end) { \
-					warnf("%s: corrupt ELF symbols", elf->filename); \
-					++sym; \
-					continue; \
-				} \
+				if (EGET(sym->st_name) >= (uint64_t)elf->len || \
+				    EGET(strtab->sh_offset) + EGET(sym->st_name) >= (uint64_t)elf->len || \
+				    !memchr(symname, 0, elf->len - EGET(strtab->sh_offset) + EGET(sym->st_name))) \
+					goto break_out; \
 				scanelf_match_symname(elf, found_sym, \
-			                          &ret, &ret_len, symname, \
-			                          ELF##B##_ST_TYPE(EGET(sym->st_info)), \
-			                          ELF##B##_ST_BIND(EGET(sym->st_info)), \
-			                          ELF##B##_ST_VISIBILITY(EGET(sym->st_other)), \
-			                          EGET(sym->st_shndx), \
-			    /* st_size can be 64bit, but no one is really that big, so screw em */ \
-			                          EGET(sym->st_size)); \
+				                      &ret, &ret_len, symname, \
+				                      ELF##B##_ST_TYPE(EGET(sym->st_info)), \
+				                      ELF##B##_ST_BIND(EGET(sym->st_info)), \
+				                      ELF##B##_ST_VISIBILITY(EGET(sym->st_other)), \
+				                      EGET(sym->st_shndx), \
+				/* st_size can be 64bit, but no one is really that big, so screw em */ \
+				                      EGET(sym->st_size)); \
 			} \
 			++sym; \
 		} \
@@ -1412,7 +1454,6 @@ static const char *scanelf_file_sym(elfobj *elf, char *found_sym)
 		FIND_SYM(64)
 	}
 
-break_out:
 	if (be_wewy_wewy_quiet) return NULL;
 
 	if (*find_sym != '*' && *found_sym)
@@ -1421,6 +1462,10 @@ break_out:
 		return NULL;
 	else
 		return " - ";
+
+ break_out:
+	warnf("%s: corrupt ELF symbols", elf->filename);
+	return NULL;
 }
 
 static const char *scanelf_file_sections(elfobj *elf, char *found_section)
@@ -2171,9 +2216,9 @@ static int parseargs(int argc, char *argv[])
 		switch (i) {
 
 		case 'V':
-			printf("pax-utils-%s: %s compiled %s\n%s\n"
+			printf("pax-utils-%s: %s\n"
 			       "%s written for Gentoo by <solar and vapier @ gentoo.org>\n",
-			       VERSION, __FILE__, __DATE__, rcsid, argv0);
+			       VERSION, VCSID, argv0);
 			exit(EXIT_SUCCESS);
 			break;
 		case 'h': usage(EXIT_SUCCESS); break;
